/*
 * BRIEF MODULE DESCRIPTION
 *	MMC framework driver for MMC/SD controllers on the
 *	Alchemy Au1100 mips processor.  Although the Au1100
 *	controllers are "Secure Digital" controllers, this
 *	module only provides a (non-secure) MultiMediaCard
 *	interface conforming to the MMC framework as developed
 *	by Andrew Christian.
 *
 *	The original MMC framework patch is located at
 *	//handhelds.org/download/linux/mmc/patch-2.4.19-mmc
 *
 *	Thanks to early users of the card whose input helped
 *	simplify things.
 *
 *	TODO: handle insertion/deletion interrupts (currently
 *	card must already be inserted before module init time).
 *
 *
 * Copyright (c) 2003 Embedded Edge, LLC.
 * Author: Embedded Edge, LLC.
 *         	dan@embeddededge.com or tim@embeddededge.com
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mmc/mmc_ll.h>
#include <asm/au1000.h>
#include <asm/au1000_dma.h>
#include <asm/au1100_mmc.h>

#if defined(CONFIG_MIPS_DB1100)
#include <asm/db1x00.h>
#elif defined(CONFIG_MIPS_EKIT1100)
#include <asm/ekit1100.h>
#else
#error "Need to import board-specific definitions for ..."
#error "... mmc_power_on() and mmc_card_inserted() macros."
#endif


/*
 *  Boot option: Force PIO - default is use DMA channels.
 *  TODO: implement boot option patchability.
 */
int au1x00_mmc_use_pio = 0;

/*
 *  Tunable quantum of PIO transfer - (11 bits available).
 */
#define AU1100_MMC_CHUNK_SIZE 128

/*
 *  Controller state flags.
 */
#define AU_MMC_STATE_INSERTED		(1<<0)
#define AU_MMC_STATE_XMIT_ACTIVE	(1<<1)
#define AU_MMC_STATE_USE_DMA		(1<<2)
#define AU_MMC_STATE_DOUBLE_BUFFER_DMA	(1<<3)
#define AU_MMC_STATE_RECV_DMA_IN_USE	(1<<4)


/*
 *  Per card/controller description.
 */
static struct au1100_mmc_card
{
	unsigned long		 m_base;	/* base of register block */
	struct mmc_request	*m_req;		/* upper level request */
	int			 m_cnt;		/* bytes processed so far */
	int			 m_flags;	/* see above flag defs */
	int			 m_tx_dma_devid;/* DMA channel identifiers */
	int			 m_rx_dma_devid;
	const char		*m_tx_dma_str;
	const char		*m_rx_dma_str;
	int			 m_xmit_dma;	/* transmit dma channel # */
	int			 m_recv_dma;	/* receive dma channel # */
	void			*m_tx_rawbuf;	/* transmit buffer */
	void			*m_rx_rawbuf;	/* receive buffer */
	dma_addr_t		 m_tx_dmaaddr;	/* virt_to_phys(buffer) */
	dma_addr_t		 m_rx_dmaaddr;
	int			 m_dma_memsize;	/* byte size of each rawbuf */
} au1100_mmc_card_table[] =
{
	{
		m_base:	SD0_BASE,
		m_flags: (AU_MMC_STATE_USE_DMA
			/* | AU_MMC_STATE_DOUBLE_BUFFER_DMA */),

		m_tx_dma_devid: DMA_NUM_DEV + DMA_ID_SD0_TX, // bank #2
		m_rx_dma_devid: DMA_NUM_DEV + DMA_ID_SD0_RX, // bank #2
		m_tx_dma_str: "MMC/SD #0 XMIT",
		m_rx_dma_str: "MMC/SD #0 RECV",
		m_xmit_dma: -1,
		m_recv_dma: -1,
	},

	{
		m_base:	SD1_BASE,
		m_flags: (AU_MMC_STATE_USE_DMA
			/* | AU_MMC_STATE_DOUBLE_BUFFER_DMA */),

		m_tx_dma_devid: DMA_NUM_DEV + DMA_ID_SD1_TX, // bank #2
		m_rx_dma_devid: DMA_NUM_DEV + DMA_ID_SD1_RX, // bank #2
		m_tx_dma_str: "MMC/SD #1 XMIT",
		m_rx_dma_str: "MMC/SD #1 RECV",
		m_xmit_dma: -1,
		m_recv_dma: -1,
	},
};

/* array for quick access - avoid runtime multiplication */
static struct au1100_mmc_card *au1100_mmc_cards[NUM_AU1100_MMC_CONTROLLERS];


/*
 *  Access macros.
 */
#define CARD_STATUS(c)	((c)->m_base + SD_STATUS)
#define CARD_CONFIG(c)	((c)->m_base + SD_CONFIG)


/*
 *  Forward declarations for MMC framework interface.
 */
static	int	au1100_mmc_init(void);
static	void	au1100_mmc_cleanup(void);
static	int	au1100_mmc_is_empty(int);
static	int	au1100_mmc_set_clock(u32);
static	void	au1100_mmc_send_cmd(struct mmc_request *);

static struct mmc_slot_driver au1100_mmc_driver =
{
	owner:		THIS_MODULE,
	name:		"Au1100 MMC Controller",
	ocr:		MMC_VDD_33_34,
	flags:		0,

	init:		au1100_mmc_init,
	cleanup:	au1100_mmc_cleanup,
	is_empty:	au1100_mmc_is_empty,
	set_clock:	au1100_mmc_set_clock,
	send_cmd:	au1100_mmc_send_cmd
};


static void au1100_mmc_send_data_pio(struct au1100_mmc_card *card)
{
	struct mmc_request *req = card->m_req;
	int		    cnt = card->m_cnt;
	int		    i = AU1100_MMC_CHUNK_SIZE;

	while (au_readl(CARD_STATUS(card)) & (SD_STATUS_TA | SD_STATUS_TH)) {

		if (cnt == req->block_len) {
			au_writel(au_readl(CARD_CONFIG(card)) & ~SD_CONFIG_TH,
				CARD_CONFIG(card));
			au_sync();
			break;
		}
		au_writel(req->buffer[cnt++], card->m_base + SD_TXPORT);
		au_sync();
		if (! i--) {
			break;
		}
	}

	DEBUG(3, ": Sent %d bytes, total so far %d\n",
		cnt - card->m_cnt, cnt);

	card->m_cnt = cnt;
}


static void au1100_mmc_receive_data_pio(struct au1100_mmc_card *card)
{
	struct mmc_request *req = card->m_req;
	int		    cnt = card->m_cnt;
	int		    i = AU1100_MMC_CHUNK_SIZE;

	while (au_readl(CARD_STATUS(card)) & SD_STATUS_NE) {
		if (cnt == req->block_len) {
			break;
		}

		req->buffer[cnt++] = au_readl(card->m_base + SD_RXPORT);
		if (! i--) {
			break;
		}
	}

	DEBUG(3, ": Rcvd %d bytes, total so far %d, total requested %d\n",
		cnt - card->m_cnt, cnt, req->block_len);

	card->m_cnt = cnt;
}


static void au1100_mmc_send_data_dma(struct au1100_mmc_card *card)
{
	struct mmc_request *req = card->m_req;
	int		    chan = card->m_xmit_dma;

	init_dma(chan);
	au_sync();

	if (card->m_flags & AU_MMC_STATE_DOUBLE_BUFFER_DMA) {
		memcpy(card->m_tx_rawbuf, req->buffer, req->block_len);
	} else {
		card->m_tx_dmaaddr = virt_to_phys(req->buffer);
	}

	if (get_dma_active_buffer(chan) == 1) {
		set_dma_count1(chan, req->block_len);
		set_dma_count0(chan, 0);
		set_dma_addr1(chan, card->m_tx_dmaaddr);
		enable_dma_buffer1(chan);
	} else {
		set_dma_count0(chan, req->block_len);
		set_dma_count1(chan, 0);
		set_dma_addr0(chan, card->m_tx_dmaaddr);
		enable_dma_buffer0(chan);
	}

	card->m_cnt = req->block_len;
	au_sync();

	if (dma_halted(chan))
		start_dma(chan);
}


static void au1100_mmc_receive_data_dma(struct au1100_mmc_card *card)
{
	struct mmc_request *req = card->m_req;
	int		    chan = card->m_recv_dma;

	if (card->m_flags & AU_MMC_STATE_RECV_DMA_IN_USE) {
		/* spurious */
		return;
	}

	card->m_flags |= AU_MMC_STATE_RECV_DMA_IN_USE;

	/*
	 *  Data is ready for reading.  Disable this interrupt
	 *  as the DMA engine now takes over.  The transfer done
	 *  interrupt will serve to let us know when data has
	 *  been moved.
	 */
	au_writel(au_readl(CARD_CONFIG(card)) & ~SD_CONFIG_RH,
		CARD_CONFIG(card));
	au_sync();

	init_dma(chan);
	au_sync();

	if ( ! (card->m_flags & AU_MMC_STATE_DOUBLE_BUFFER_DMA) ) {
		card->m_rx_dmaaddr = virt_to_phys(req->buffer);
	}

	if (get_dma_active_buffer(chan) == 1) {
		set_dma_count1(chan, req->block_len);
		set_dma_count0(chan, 0);
		set_dma_addr1(chan, card->m_rx_dmaaddr);
		enable_dma_buffer1(chan);
	} else {
		set_dma_count0(chan, req->block_len);
		set_dma_count1(chan, 0);
		set_dma_addr0(chan, card->m_rx_dmaaddr);
		enable_dma_buffer0(chan);
	}

	card->m_cnt = req->block_len;
	au_sync();

	if (dma_halted(chan))
		start_dma(chan);
}


static void au1100_mmc_dma_complete(struct au1100_mmc_card *card)
{
	unsigned int buf_done;
	int chan;

	if (card->m_flags & AU_MMC_STATE_XMIT_ACTIVE)
		chan = card->m_xmit_dma;
	else
		chan = card->m_recv_dma;

	/*
	 *  On receive, MMC may signal xfer is done before dma
	 *  engine has finished reading the fifo and actually
	 *  transfering data to memory.  A small window perhaps,
	 *  but we wait until the engine says it's through too.
	 */
	while ((buf_done = get_dma_buffer_done(chan)) == 0) {
		DEBUG(1, ": waiting for MMC DMA completion.\n");

		au_sync();
	}

	if (buf_done == (DMA_D0 | DMA_D1)) {
		DEBUG(0, ": MMC DMA error - multiple outstanding buffers\n");

		card->m_cnt = 0;	// signal error
	}

	disable_dma(chan);

	if (card->m_flags & AU_MMC_STATE_RECV_DMA_IN_USE) {
		card->m_flags &= ~AU_MMC_STATE_RECV_DMA_IN_USE;

		if (card->m_flags & AU_MMC_STATE_DOUBLE_BUFFER_DMA) {
			memcpy( card->m_req->buffer,
				card->m_rx_rawbuf,
				card->m_cnt);
		}

		/* reenable for next time */
		au_writel(au_readl(CARD_CONFIG(card)) | SD_CONFIG_RH,
			CARD_CONFIG(card));
		au_sync();
	}
}


static void au1100_mmc_cmd_ack(struct au1100_mmc_card *card)
{
	struct mmc_request *req = card->m_req;
	int base = card->m_base;

	switch (req->rtype) {

	case RESPONSE_NONE:
		break;

	case RESPONSE_R3:
	case RESPONSE_R1:
	case RESPONSE_R1B:
	  {
		u8 header = au_readl(base + SD_RESP1);
		u32 ocr = au_readl(base + SD_RESP0);

		req->response[0] = header;
		req->response[1] = (ocr >> 24) & 0xff;
		req->response[2] = (ocr >> 16) & 0xff;
		req->response[3] = (ocr >>  8) & 0xff;
		req->response[4] = (ocr) & 0xff;
		break;
	  }

	case RESPONSE_R2_CID:
	case RESPONSE_R2_CSD:
	  {
		u32 reg[4];
		int i;

		reg[0] = au_readl(base + SD_RESP3);
		reg[1] = au_readl(base + SD_RESP2);
		reg[2] = au_readl(base + SD_RESP1);
		reg[3] = au_readl(base + SD_RESP0);

		for (i = 0; i < 4; i++) {
			if (i) {
				req->response[i*4] = (reg[i] >> 24) & 0xff;
			}
			req->response[i*4 + 1] = (reg[i] >> 16) & 0xff;
			req->response[i*4 + 2] = (reg[i] >> 8) & 0xff;
			req->response[i*4 + 3] = (reg[i]) & 0xff;
		}
		req->response[0] = 0x3f;
		break;
	  }

	case RESPONSE_R4:
	case RESPONSE_R5:
	default:
		DEBUG(0, ": Unsupported MMC response type %d\n", req->rtype);
		req->result = MMC_ERROR_DRIVER_FAILURE;
		break;
	}

	req->result = MMC_NO_ERROR;
	if (!req->nob) {
		mmc_cmd_complete(req);
	}
}


static void au1100_mmc_irq(int irq, void *dev_id, struct pt_regs *regs)
{
	int mmc;

	/*
	 *  Traverse all irq sources.
	 */
	for (mmc = 0; mmc < NUM_AU1100_MMC_CONTROLLERS; ++mmc) {

		struct au1100_mmc_card *card = au1100_mmc_cards[mmc];

		int tx = card->m_flags & AU_MMC_STATE_XMIT_ACTIVE;
		int dma = card->m_flags & AU_MMC_STATE_USE_DMA;

		u32 status = au_readl(CARD_STATUS(card));

		DEBUG(2, ": card %d status 0x%08X\n", mmc, status);

		if (status & SD_STATUS_SI) {
			/* TODO: FIX */
			BUG();
		}
		if (status & SD_STATUS_CD) {
			/* TODO: FIX */
			BUG();
		}

		if (status & SD_STATUS_CR) {
			au1100_mmc_cmd_ack(card);

			if (tx) {
				if (dma) {
					au1100_mmc_send_data_dma(card);
				} else {
					au_writel(au_readl(CARD_CONFIG(card))
							| SD_CONFIG_TH,
						CARD_CONFIG(card));
					au_sync();
					if (status & ( SD_STATUS_TH
						     | SD_STATUS_TA
						     | SD_STATUS_TE)) {
						au1100_mmc_send_data_pio(card);
					}
				}
			} else {
				/*
				 *  Optimization.  Re-read status to possibly
				 *  catch RX Data-Ready instead of incurring
				 *  additional interrupt to tell us so.
				 */
				u32 status2;
				au_sync();
				status2 = au_readl(CARD_STATUS(card));
				DEBUG(2, ": card %d status2 0x%08X\n",
					mmc, status2);

				status |= status2;
			}
		} else if (status & (SD_STATUS_TA | SD_STATUS_TE)) {
			if (tx && !dma) {
				au1100_mmc_send_data_pio(card);
			}
		}

		if (status & SD_STATUS_DD) {
			if (dma) {
				au1100_mmc_dma_complete(card);
			}

			if (card->m_req->block_len != card->m_cnt) {
				card->m_req->result = MMC_ERROR_CRC;
				DEBUG(0, ": Incomplete transfer %s - 0x%08x\n",
					tx ? "writing" : "reading",
					status);
			}
			mmc_cmd_complete(card->m_req);
			card->m_req = 0;
			card->m_cnt = 0;
			if (tx) {
				card->m_flags &= ~AU_MMC_STATE_XMIT_ACTIVE;

				if (! dma ) {
					au_writel(au_readl(CARD_CONFIG(card))
							& ~SD_CONFIG_TH,
						CARD_CONFIG(card));
					au_sync();
				}
			}
		}

		/*
		 *  Depending on timing, x_NE and x_DD may
		 *  both be asserted momentarily in the status
		 *  word, thus we test to see if we still have
		 *  an outstanding request.
		 */
		if ((card->m_req) && (status & SD_STATUS_NE)) {
			if (dma) {
				au1100_mmc_receive_data_dma(card);
			} else {
				au1100_mmc_receive_data_pio(card);
			}
		}

		au_writel(status, CARD_STATUS(card));
		au_sync();
	}
}


static void au1100_mmc_send_cmd(struct mmc_request *req)
{
	int mmc = req->index;
	struct au1100_mmc_card *card = au1100_mmc_cards[mmc];
	u32 mmc_cmd = 0;

	DEBUG(2, ": card (%d) command = %d\n", mmc, req->cmd);

	/*
	 *  Establish command type.
	 */
	switch (req->cmd) {

	case MMC_CIM_RESET:
		card->m_flags &= ~AU_MMC_STATE_XMIT_ACTIVE;
		card->m_req = 0;
		card->m_cnt = 0;

		req->result = MMC_NO_ERROR;
		mmc_cmd_complete(req);
		return;

	case MMC_WRITE_BLOCK:
		mmc_cmd |= SD_CMD_CT_1;
		break;

	case MMC_READ_SINGLE_BLOCK:
		mmc_cmd |= SD_CMD_CT_2;
		break;

	case MMC_WRITE_MULTIPLE_BLOCK:	/* SD_CMD_CT_3 */
	case MMC_READ_MULTIPLE_BLOCK:	/* SD_CMD_CT_4 */
	case MMC_STOP_TRANSMISSION:	/* SD_CMD_CT_7 */
	  {
		u32 ctype=0;
		switch (req->cmd) {
		case MMC_WRITE_MULTIPLE_BLOCK:	ctype = SD_CMD_CT_3; break;
		case MMC_READ_MULTIPLE_BLOCK:	ctype = SD_CMD_CT_4; break;
		case MMC_STOP_TRANSMISSION:	ctype = SD_CMD_CT_7; break;
		}
		DEBUG(0, ": Unsupported MMC command type: %d\n", ctype);

		req->result = MMC_ERROR_ILLEGAL_COMMAND;
		mmc_cmd_complete(req);
		return;
	  }

	default:
		if (req->nob) {
			DEBUG(0, ": Data blocks (%d) for non-I/O command %d\n",
				req->nob, req->cmd);

			req->result = MMC_ERROR_DRIVER_FAILURE;
			mmc_cmd_complete(req);
			return;
		}
		break;
	}

	/*
	 *  Insert upper level command.
	 */
	mmc_cmd |= req->cmd << SD_CMD_CI_SHIFT;

	/*
	 *  Deal with only one data block at a time.
	 */
	if (req->nob && req->nob != 1) {
		DEBUG(0, ": Too many data blocks (%d) for command %d\n",
			req->nob, req->cmd);

		req->result = MMC_ERROR_DRIVER_FAILURE;
		mmc_cmd_complete(req);
		return;
	}

	/*
	 *  Establish response type.
	 */
	switch (req->rtype) {

	case RESPONSE_NONE:
		mmc_cmd |= SD_CMD_RT_0;
		break;

	case RESPONSE_R1:
		mmc_cmd |= SD_CMD_RT_1;
		break;

	case RESPONSE_R1B:
		mmc_cmd |= SD_CMD_RT_1B;
		break;

	case RESPONSE_R2_CID:
	case RESPONSE_R2_CSD:
		mmc_cmd |= SD_CMD_RT_2;
		break;

	case RESPONSE_R3:
		mmc_cmd |= SD_CMD_RT_3;
		break;

	case RESPONSE_R4:
	case RESPONSE_R5:
	default:
		DEBUG(0, ": Unsupported MMC response type %d\n", req->rtype);

		req->result = MMC_ERROR_ILLEGAL_COMMAND;
		mmc_cmd_complete(req);
		return;
	}

	/*
	 *  Command argument must be written first.
	 */
	au_writel(req->arg, card->m_base + SD_CMDARG);

	/*
	 *  Block count and size next.
	 */
	if (req->nob) {
		/*
		 *  Value programmed is one less than the actual
		 *  number of bytes in the block.
		 */
		au_sync();	/* ordered writes */
		au_writel(((req->block_len-1) << SD_BLKSIZE_BS_SHIFT)
				  | (req->nob << SD_BLKSIZE_BC_SHIFT),
			card->m_base + SD_BLKSIZE);
	}

	card->m_req = req;
	card->m_cnt = 0;

	if (req->cmd == MMC_WRITE_BLOCK) {
		card->m_flags |= AU_MMC_STATE_XMIT_ACTIVE;
	} else {
		card->m_flags &= ~AU_MMC_STATE_XMIT_ACTIVE;
	}

	/*
	 *  Make sure state is updated before starting command.
	 */
	au_sync();	/* ordered writes */
	au_writel((mmc_cmd | SD_CMD_GO), card->m_base + SD_CMD);
}


static int au1100_mmc_is_empty(int mmc)
{
	int ins=0;

	mmc_card_inserted(mmc, &ins);

	/*
	 *  Update our own sense of presence.
	 */
	if (ins)
		au1100_mmc_cards[mmc]->m_flags |= AU_MMC_STATE_INSERTED;
	else
		au1100_mmc_cards[mmc]->m_flags &= ~AU_MMC_STATE_INSERTED;

	/*
	 *  TODO: should we attempt to allocate DMA resources here??
	 */

	return ( ! ins );
}


static int au1100_mmc_set_clock(u32 rate)
{
	/* TODO: set clock */
	DEBUG(1, ": rate=%d(0x%x)\n", rate, rate);
	return 0;
}


static int au1100_mmc_init(void)
{
	int mmc, ret;

	/*
	 *  Get irq.  Last arg unused as service routine will scan
	 *  all cards.
	 */
	ret = request_irq(AU1100_SD_IRQ, au1100_mmc_irq, SA_SHIRQ,
				"SD", au1100_mmc_cards[0]/* unused */);
	if (ret) {
		DEBUG(0, ": request_irq() failed %d(0x%x).\n", -ret, -ret);
		return ret;
	}

	for (mmc = 0; mmc < NUM_AU1100_MMC_CONTROLLERS; ++mmc) {

		struct au1100_mmc_card *card = au1100_mmc_cards[mmc];
		int ins=0;

		/*
		 *  Process boot option
		 */
		if ( au1x00_mmc_use_pio )
			card->m_flags &= ~AU_MMC_STATE_USE_DMA;
		else
			card->m_flags |= AU_MMC_STATE_USE_DMA;

		/*
		 *  Enable controller via reset sequence.
		 *  Data book says first enable clock, then reset peripheral.
		 */
		au_writel(SD_ENABLE_CE, card->m_base + SD_ENABLE);
		au_sync();	/* ordered writes */
		au_writel((SD_ENABLE_R | SD_ENABLE_CE),card->m_base+SD_ENABLE);
		au_sync();	/* ordered writes */

#ifdef CONFIG_MIPS_EKIT1100
		/*
		 * Frequency of SD controller clock is set to 4.95MHz(=99MHz/2/(9+1)).
		 */
		au_writel((au_readl(card->m_base + SD_CONFIG) & ~SD_CONFIG_DIV)
				 | 9 | SD_CONFIG_DE, card->m_base + SD_CONFIG);
		au_sync();
#endif	
		/*
		 *  Enable serial interface state machine and FIFO master.
		 */
		au_writel(SD_CONFIG2_EN, card->m_base + SD_CONFIG2);
		au_sync();	/* ordered writes */

		/*
		 *  Set max timeout - 21 bits - rest are reserved.
		 */
		au_writel(0x1fffff, card->m_base + SD_TIMEOUT);
		au_sync();	/* ordered writes */

		/*
		 *  Detect card.
		 */
		mmc_card_inserted(mmc, &ins);
		if (ins)
			card->m_flags |= AU_MMC_STATE_INSERTED;
		else
			card->m_flags &= ~AU_MMC_STATE_INSERTED;

		/*
		 *  Clear status bits after detection in case
		 *  the process of detection altered them.
		 */
		au_writel(au_readl(CARD_STATUS(card)), CARD_STATUS(card));
		au_sync();	/* ordered writes */

		/*
		 *  Enable interrupts.
		 */
		au_writel((SD_CONFIG_I |	/* master enable */
			   SD_CONFIG_CR |	/* command response */
			   SD_CONFIG_DD |	/* xfer done */
			   SD_CONFIG_RH),	/* recv ready */
			CARD_CONFIG(card));
		au_sync();	/* ordered writes */

		/*
		 *  Attach DMA resources.
		 */
		ret = -1;
		if ( (card->m_flags & AU_MMC_STATE_INSERTED)
		  && (card->m_flags & AU_MMC_STATE_USE_DMA) ) {

			DEBUG(1, ": allocating TX DMA chan for MMC #%d\n", mmc);

			ret = card->m_xmit_dma =
				request_au1000_dma(card->m_tx_dma_devid,
						card->m_tx_dma_str,
						NULL, 0, NULL);

			if (ret < 0) {
				DEBUG(0, ": Request %s DMA failed %d(0x%x).\n",
					card->m_tx_dma_str,
					-ret, -ret);
			}
		}

		if (ret >= 0) {

			DEBUG(1, ": allocating RX DMA chan for MMC #%d\n", mmc);

			ret = card->m_recv_dma =
				request_au1000_dma(card->m_rx_dma_devid,
						card->m_rx_dma_str,
						NULL, 0, NULL);

			if (ret < 0) {
				DEBUG(0, ": Request %s DMA failed %d(0x%x).\n",
					card->m_rx_dma_str,
					-ret, -ret);

				/* undo xmit allocation */
				if (card->m_xmit_dma >= 0) {
					free_au1000_dma(card->m_xmit_dma);
					card->m_xmit_dma = -1;
				}
			}
		}

		/* Get some memory for DMA. */
		if (ret >= 0) {
			void *mem1, *mem2;
			int memsize;

			/* maximum size (bytes) of MMC transaction */
			memsize = (SD_BLKSIZE_BS >> SD_BLKSIZE_BS_SHIFT) + 1;

			/* round maximum bytes up to next page boundary */
			memsize = (memsize + PAGE_SIZE - 1) & PAGE_MASK;

			mem1 = (void *)__get_free_pages(GFP_ATOMIC | GFP_DMA,
						     get_order(memsize));
			mem2 = (void *)__get_free_pages(GFP_ATOMIC | GFP_DMA,
						     get_order(memsize));
			if (mem1 != NULL && mem2 != NULL) {
				memset(mem1, 0, memsize);
				memset(mem2, 0, memsize);
				card->m_tx_rawbuf = mem1;
				card->m_rx_rawbuf = mem2;
				card->m_tx_dmaaddr = virt_to_phys(mem1);
				card->m_rx_dmaaddr = virt_to_phys(mem2);

				card->m_dma_memsize = memsize;
			} else {
				ret = -1;

				/*
				 *  Release resources and fall back to PIO.
				 */
				DEBUG(0, ": DMA buffer allocation(s) failed"
					" (0x%x 0x%x).\n",
					(unsigned int)mem1,
					(unsigned int)mem2);

				if (mem1 != NULL) {
					free_pages((unsigned long)mem1,
						get_order(memsize));
				}
				if (mem2 != NULL) {
					free_pages((unsigned long)mem2,
						get_order(memsize));
				}
				free_au1000_dma(card->m_xmit_dma);
				card->m_xmit_dma = -1;
				free_au1000_dma(card->m_recv_dma);
				card->m_recv_dma = -1;
			}
		}

		/*
		 *  Fall back to PIO if DMA setup fails.
		 */
		if ( (card->m_flags & AU_MMC_STATE_INSERTED)
		  && (card->m_flags & AU_MMC_STATE_USE_DMA)
		  && (ret < 0) ) {
			card->m_flags &= ~AU_MMC_STATE_USE_DMA;

			DEBUG(0,": DMA setup for MMC #%d failed - using PIO\n",
				mmc);
		}
	}
	return 0;
}


static void au1100_mmc_cleanup(void)
{
	int mmc;
	struct au1100_mmc_card *card;

	/*
	 *  Disable peripherals.
	 */
	au_sync();	/* ordered writes */
	for (mmc = 0; mmc < NUM_AU1100_MMC_CONTROLLERS; ++mmc) {
		card = au1100_mmc_cards[mmc];

		au_writel(0, card->m_base + SD_ENABLE);
	}
	au_sync();	/* ordered writes */

	/*
	 *  Release DMA resources.
	 */
	for (mmc = 0; mmc < NUM_AU1100_MMC_CONTROLLERS; ++mmc) {
		card = au1100_mmc_cards[mmc];

		if (card->m_tx_rawbuf != NULL) {
			free_pages((unsigned long)card->m_tx_rawbuf,
				get_order(card->m_dma_memsize));
			card->m_tx_rawbuf = NULL;
		}
		if (card->m_rx_rawbuf != NULL) {
			free_pages((unsigned long)card->m_rx_rawbuf,
				get_order(card->m_dma_memsize));
			card->m_rx_rawbuf = NULL;
		}
		if (card->m_xmit_dma >= 0) {
			free_au1000_dma(card->m_xmit_dma);
			card->m_xmit_dma = -1;
		}
		if (card->m_recv_dma >= 0) {
			free_au1000_dma(card->m_recv_dma);
			card->m_recv_dma = -1;
		}
	}

	free_irq(AU1100_SD_IRQ, au1100_mmc_cards[0]);
}


static int __init au1100_mmc_module_init(void)
{
	int mmc;

	for (mmc = 0; mmc < NUM_AU1100_MMC_CONTROLLERS; ++mmc) {

		mmc_power_on(mmc);

		DEBUG(2, ": powering MMC/SD card #%d\n", mmc);

		au1100_mmc_cards[mmc] = &au1100_mmc_card_table[mmc];
	}

	/*
	 *  Leave peripheral clock disabled to conserve power.
	 */
	au_sync();	/* ordered writes */
	for (mmc = 0; mmc < NUM_AU1100_MMC_CONTROLLERS; ++mmc) {

		au_writel(0, au1100_mmc_cards[mmc]->m_base + SD_ENABLE);
	}
	au_sync();	/* ordered writes */

	/*
	 *  Insert self into MMC framework.
	 */
	return mmc_register_slot_driver(&au1100_mmc_driver,
		NUM_AU1100_MMC_CONTROLLERS);
}


static void __exit au1100_mmc_module_exit(void)
{
	mmc_unregister_slot_driver(&au1100_mmc_driver);
}


module_init(au1100_mmc_module_init);
module_exit(au1100_mmc_module_exit);


MODULE_AUTHOR("Embedded Edge, LLC.");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Au1100 MultiMediaCard controller support");


